#include <math.h>
#include <string.h>
#include <stdio.h>
#include <SDL.h>
#include <SDL_audio.h>

#include "Constants.h"
#include "Timing.h"
#include "Log.h"
#include "Sound.h"
#include "SoundBuffer.h"

static Uint32 _sampleRatePerSecond;

const Uint8 NUMBER_OF_CHANNELS = 2; // stereo

const Sint16 AUDIO_HIGH_I = 0x3DF0;
const Sint16 AUDIO_LOW_I = 0x0000;

const float AUDIO_HIGH_F = 0.10f;
const float AUDIO_LOW_F = 0.0f;

static Uint8 _soundInitialized = 0;
enum soundSamplingMode _samplingMode;
enum soundMode _userRequestedSoundMode;

//    ++-----------------------sample is signed if set
//    ||
//    ||       ++-----------sample is bigendian if set
//    ||       ||
//    ||       ||          ++---sample is float if set
//    ||       ||          ||
//    ||       ||          ||     +---sample bit size---+
//    ||       ||          ||     |                     |
//    15 14 13 12 11 10 09 08     07 06 05 04 03 02 01 00
SDL_AudioFormat _format;
Uint8 _formatBitsPerSample;
Uint8 _formatIsFloat;

// used to decide when to sample the speaker, writing into the sound buffer
static Uint32 _spectrumSpeakerSamplingIntervalTstates;
static Uint64 _tstatesAtLastSample;

static Uint64 _frame = 0;

#define _DEBUG_BUFFER_SIZE 256
static const char* _debugBuffer[_DEBUG_BUFFER_SIZE];

// the write operation treats the buffer as linear, to not overwrite earlier samples
void _sound_spectrum_buffer_write(Uint8 value8) {
    if (!_soundInitialized) {
        return;
    }

    soundbuffer_write(value8);
    return;
}

Uint8 _sound_spectrum_buffer_read() {
    if (!_soundInitialized) {
        return 0xF3;
    }

    return soundbuffer_read();
}

void sound_play() {
    if (!_soundInitialized) {
        return;
    }

    SDL_PauseAudio(0);      // start playing sound
}

void sound_set_user_mode(enum soundMode mode) {
    _userRequestedSoundMode = mode;
    
    switch (_userRequestedSoundMode)
    {
    case Auto:
        break;
    case Static:
        // force it immediately
        sound_set_sampling_mode(SoundStaticSampling);
        break;
    case Dynamic:
        // force it immediately
        sound_set_sampling_mode(SoundDynamicSampling);
        break;
    default:
        break;
    }
}

void sound_set_sampling_mode(enum soundSamplingMode mode) {
    if (_userRequestedSoundMode == Static && mode == SoundDynamicSampling) {
        // mode forced by user and requested mode do not match
        return;
    }

    if (_userRequestedSoundMode == Dynamic && mode == SoundStaticSampling) {
        // mode forced by user and requested mode do not match
        return;
    }

    _samplingMode = mode;
}

// determines sampling interval
void _sound_compute_write_timing() {
    double soundFrameDurationMs;

    // this is valuable only when frames are allowed to go over their target run time by
    // a few percent
    if (_samplingMode == SoundStaticSampling ) {
        // when frames are at their target run time, static sampling means that the target
        // is used to compute speaker sampling interval
        soundFrameDurationMs = timing_get_target_milliseconds_per_video_frame();
    }
    else {
        // when frames are a few percent slower than their target run time, the sound quality
        // is improved by computing a speaker sampling interval based on an actual frame
        // run time average
        soundFrameDurationMs = timing_get_actual_milliseconds_per_video_frame();
    }

    Uint32 _samplesPerFrame = (Uint32)((double)_sampleRatePerSecond * (soundFrameDurationMs / 1000.0f));
    // calculate sampling interval in tstates
    Uint64 tstatesPerVideoFrame = timing_get_states_per_video_frame();
    _spectrumSpeakerSamplingIntervalTstates = (Uint32)((double)tstatesPerVideoFrame / (double)_samplesPerFrame);
}

void _sound_buffer_filler_callback(void* consumerData, Uint8* sdlBuffer, int bytesToFill) {
    int lengthInSamples = bytesToFill / (_formatBitsPerSample/8);
    lengthInSamples = lengthInSamples / NUMBER_OF_CHANNELS;

    if (_formatIsFloat) {
        // samples must be floating point

        if (_formatBitsPerSample == 32) {
            float* buffer = (float*)sdlBuffer;
            for (int i = 0; i < lengthInSamples*NUMBER_OF_CHANNELS; i+=NUMBER_OF_CHANNELS) {
                float value = _sound_spectrum_buffer_read() ? AUDIO_HIGH_F : AUDIO_LOW_F;
                for (int j = 0; j < NUMBER_OF_CHANNELS; j++) {
                    buffer[i + j] = value;
                }
            }
        }
        else if (_formatBitsPerSample == 64) {
            double* buffer = (double*)sdlBuffer;
            for (int i = 0; i < lengthInSamples * NUMBER_OF_CHANNELS; i += NUMBER_OF_CHANNELS) {
                double value = _sound_spectrum_buffer_read() ? (double)AUDIO_HIGH_F : (double)AUDIO_LOW_F;
                for (int j = 0; j < NUMBER_OF_CHANNELS; j++) {
                    buffer[i + j] = value;
                }
            }
        }
    }
    else {
        // samples must be integers

        if (_formatBitsPerSample == 16) {
            Uint16* buffer = (Uint16*)sdlBuffer;
            for (int i = 0; i < lengthInSamples * NUMBER_OF_CHANNELS; i += NUMBER_OF_CHANNELS) {
                Uint16 value = _sound_spectrum_buffer_read() ? (Uint16)AUDIO_HIGH_I : (Uint16)AUDIO_LOW_I;
                for (int j = 0; j < NUMBER_OF_CHANNELS; j++) {
                    buffer[i + j] = value;
                }
            }
        }
        else if (_formatBitsPerSample == 32) {
            Uint32* buffer = (Uint32*)sdlBuffer;
            for (int i = 0; i < lengthInSamples * NUMBER_OF_CHANNELS; i += NUMBER_OF_CHANNELS) {
                Uint32 value = _sound_spectrum_buffer_read() ? (Uint32)AUDIO_HIGH_I : (Uint32)AUDIO_LOW_I;
                for (int j = 0; j < NUMBER_OF_CHANNELS; j++) {
                    buffer[i + j] = value;
                }
            }
        }
    }
}

// called after each CPU instruction, it is in charge of deciding
// if it's time to write to the sound buffer
void sound_add_sample(Uint8 isSpeakerActive, Uint64 totalTStates) {
    if (!_soundInitialized) {
        return;
    }

    if (totalTStates - _tstatesAtLastSample >= _spectrumSpeakerSamplingIntervalTstates) {
        // this allows for a catchup on consecutive CPU instructions, if, for some reason,
        // we fall behind for more than one sampling interval
        _tstatesAtLastSample += _spectrumSpeakerSamplingIntervalTstates;

        if (timing_is_faster()) {
            // silence during faster CPU - such as when loading
            _sound_spectrum_buffer_write(0);
        }
        else {
            _sound_spectrum_buffer_write(isSpeakerActive);
        }
    }
}

static Uint64 _framesWithStaticSound = 0;
static Uint64 _framesWithDynamicSound = 0;

void sound_notify_end_of_frame() {
    if (!_soundInitialized) {
        return;
    }

    _frame++;
    _sound_compute_write_timing();
    if (_samplingMode == SoundStaticSampling) {
        _framesWithStaticSound++;
    }
    else {
        _framesWithDynamicSound++;
    }
}

void sound_destroy() {
    if (!_soundInitialized) {
        return;
    }

    _soundInitialized = 0;

    SDL_PauseAudio(1); // stop playing sound
    SDL_CloseAudio();

    log_write("SOUND");

    switch (_userRequestedSoundMode)
    {
    case Auto:
        log_write("Sound mode: Auto");
        break;
    case Static:
        log_write("Sound mode: Static");
        break;
    case Dynamic:
        log_write("Sound mode: Dynamic");
        break;
    default:
        break;
    }

    char* message = (char*)malloc(500 * sizeof(char));
    if (message) {
        sprintf_s(message, 490, "Frame counts by sound timing type: %I64d static, %I64d dynamic",
            _framesWithStaticSound,
            _framesWithDynamicSound);
        log_write(message);
        free(message);
    }

    soundbuffer_destroy();
}

void sound_start(Uint32 samplesPerSecond) {
    sound_destroy();

    _sampleRatePerSecond = samplesPerSecond;

    if (SDL_Init(SDL_INIT_AUDIO) != 0) {
        log_write("Error: could not initialize SDL audio");
        log_write((char*)SDL_GetError());
        return;
    }

    sound_set_sampling_mode(SoundStaticSampling);

    SDL_AudioSpec want;
    want.freq = _sampleRatePerSecond;
    want.format = AUDIO_F32SYS;
    want.channels = NUMBER_OF_CHANNELS;
    want.samples = 0x1000;  // buffer size
    want.callback = _sound_buffer_filler_callback;
    want.userdata = NULL;   // unused
    want.silence = 0;

    char* message = (char*)malloc(500 * sizeof(char));
    if (message) {
        sprintf_s(message, 490, "Requesting audio spec %04x", want.format);
        log_write(message);
        free(message);
    }

    SDL_AudioSpec got;
    if (SDL_OpenAudio(&want, &got) != 0) {
        log_write("Error: could not open audio");
        log_write((char*)SDL_GetError());
        return;
    }
    _format = got.format;
    _formatBitsPerSample = SDL_AUDIO_BITSIZE(got.format);
    _formatIsFloat = SDL_AUDIO_ISFLOAT(got.format) ? 1 : 0;

    message = (char*)malloc(500 * sizeof(char));
    if (message) {
        sprintf_s(message, 490, "Got audio spec %04x", got.format);
        log_write(message);
        free(message);
    }

    if (SDL_AUDIO_ISBIGENDIAN(got.format)) {
        log_write("Warning: got audio spec that is big endian, which is unsupported; audio will likely be silence or garbage");
    }

    if (want.format != got.format || 
        want.channels != got.channels ||
        want.freq != got.freq) {

        log_write("Warning: got an audio spec different than the one requested");
    }

    _sound_compute_write_timing();
    _tstatesAtLastSample = 0;

    soundbuffer_start(_sampleRatePerSecond);
    _soundInitialized = 1;
}
