#include <stdio.h>
#include <string.h>

#include "SoundBuffer.h"
#include "Constants.h"
#include "Timing.h"
#include "Log.h"

// holds our samples, one byte per sample
static Uint8* _buffer;
static Uint32 _bufferSize;

// tends to take longer to synchronize on faster computers/newer OSes
// so we wait for this long before we record any resyncs, for the statistics
// EXPERIMENTALLY DETERMINED
static double _initialSyncPeriodSeconds = 5.0f;

static Uint32 _sampleRatePerSecond;

// these represent the read and write "heads"
//     SDL's audio module reads
//     CPU activity writes
static Uint32 _read_index;
static Uint32 _write_index;

// tracks total distance travelled by read and write "heads"
static Sint64 _read_index_distance_traveled;
static Sint64 _write_index_distance_traveled;

// counts all occurrences when resynchronization is needed, so it can be reported in the log
static Uint32 _resync_count__read_is_too_close_behind_write;
static Uint32 _resync_count__read_is_too_far_behind_write;

#define _DEBUG_BUFFER_SIZE 256
static const char* _debugBuffer[_DEBUG_BUFFER_SIZE];

static Uint8 _resyncLoggingEnabled = 0;

// resynchronization parameters
Uint32 _read_minimum_trailing_distance;
Uint32 _read_maximum_trailing_distance;
Uint32 _read_sync_rewind_amount;
Uint32 _read_sync_fastforward_amount;

#define DEFAULT_READ_MINIMUM_TRAILING_DISTANCE__FACTOR 100
#define DEFAULT_READ_MAXIMUM_TRAILING_DISTANCE__FACTOR  1500
// amounts above 100 cause constant stutter in manic miner
#define DEFAULT_READ_SYNC_REWIND_AMOUNT__FACTOR         50
#define DEFAULT_READ_SYNC_FASTFORWARD_AMOUNT__FACTOR    50

// multiplied by the above to allow for integer-based fine adjustments
static Uint32 _sync_divisor;

// EXPERIMENTALLY DETERMINED
static double _bufferSizeSeconds = 0.33333333333f;
void soundbuffer_set_buffer_size(double seconds) {
    _bufferSizeSeconds = seconds;
}

void _compute_sync_divisor() {
    // all values in this function are determined experimentally
    if (timing_get_target_milliseconds_per_video_frame() <= 20) {
        // this has been the default ever since sound sync was implemented
        _sync_divisor = 30000;
    }
    else {
        // I discovered that this greatly improved sound quality for
        // when the CPU was slowed down (via high -framems settings)
        // in most cases, it reduced the number of resyncs 10-fold
        _sync_divisor = 8000;
    }
}

static Uint16 _sync__min_trailing_distance_factor = DEFAULT_READ_MINIMUM_TRAILING_DISTANCE__FACTOR;
Uint16 _sync__get_min_trailing_distance_factor() {
    return _sync__min_trailing_distance_factor;
}
void soundbuffer_sync__set_min_trailing_distance_factor(Uint16 factor) {
    _sync__min_trailing_distance_factor = factor;
}

static Uint16 _sync__max_trailing_distance_factor = DEFAULT_READ_MAXIMUM_TRAILING_DISTANCE__FACTOR;
Uint16 _sync__get_max_trailing_distance_factor() {
    return _sync__max_trailing_distance_factor;
}
void soundbuffer_sync__set_max_trailing_distance_factor(Uint16 factor) {
    _sync__max_trailing_distance_factor = factor;
}

static Uint16 _sync__rewind_amount_factor = DEFAULT_READ_SYNC_REWIND_AMOUNT__FACTOR;
Uint16 _sync__get_rewind_amount_factor() {
    return _sync__rewind_amount_factor;
}
void soundbuffer_sync__set_rewind_amount_factor(Uint16 factor) {
    _sync__rewind_amount_factor = factor;
}

static Uint16 _sync__fastforward_amount_factor = DEFAULT_READ_SYNC_FASTFORWARD_AMOUNT__FACTOR;
Uint16 _sync__get_fastforward_amount_factor() {
    return _sync__fastforward_amount_factor;
}
void soundbuffer_sync__set_fastforward_amount_factor(Uint16 factor) {
    _sync__fastforward_amount_factor = factor;
}

void _compute_sync_values() {
    _read_minimum_trailing_distance = _sampleRatePerSecond * _sync__get_min_trailing_distance_factor() / _sync_divisor;
    _read_maximum_trailing_distance = _sampleRatePerSecond * _sync__get_max_trailing_distance_factor() / _sync_divisor;
    _read_sync_rewind_amount = _sampleRatePerSecond * _sync__get_rewind_amount_factor() / _sync_divisor;
    _read_sync_fastforward_amount = _sampleRatePerSecond * _sync__get_fastforward_amount_factor() / _sync_divisor;
}

// rewinds (moves "backward") a circular buffer index by AT MOST one full buffer length
void _soundbuffer_rewind_index(Uint32* index, Uint64* indexDistanceTravelled, Uint32 rewindAmount) {
    if (*index >= rewindAmount) {
        *index -= rewindAmount;
    }
    else {
        *index = *index + _bufferSize - rewindAmount;
    }

    *indexDistanceTravelled -= rewindAmount;
}

// fast-forwards (moves "forward") a circular buffer index by AT MOST one full buffer length
void _soundbuffer_fastforward_index(Uint32* index, Uint64* indexDistanceTravelled, Uint32 fastForwardAmount) {
    *index += fastForwardAmount;
    if (*index >= _bufferSize) {
        *index -= _bufferSize;
    }

    *indexDistanceTravelled += fastForwardAmount;
}

void _soundbuffer_rewind_read() {
    _soundbuffer_rewind_index(&_read_index, &_read_index_distance_traveled, _read_sync_rewind_amount);
}

void _soundbuffer_fastforward_read() {
    _soundbuffer_fastforward_index(&_read_index, &_read_index_distance_traveled, _read_sync_fastforward_amount);
}

// depending on computer and OS, sound is poorly synchronized for a short while at application start
// this can report much higher resync numbers
Uint8 _soundbuffer_is_in_initial_sync_period() {
    Uint32 syncPeriodLengthInSamples = (Uint32)((double)_sampleRatePerSecond * _initialSyncPeriodSeconds);
    if (_read_index_distance_traveled > syncPeriodLengthInSamples) {
        // we've read past the initial sync period
        return 0;
    }

    return 1;
}

Uint8 soundbuffer_read() {
    Sint64 distance = _write_index_distance_traveled - _read_index_distance_traveled;
    // synchronize when either:
    //     - read has almost caught up to or is ahead of write, or
    //     - read is too far behind write
    if (distance < (Sint64)_read_minimum_trailing_distance) {
        // read has almost caught up to write (read is too fast)
        if (!timing_is_faster()) {
            // only gather statistics for when CPU is running at normal speed
            if (!_soundbuffer_is_in_initial_sync_period()) {
                // and not in the initial sync period
                _resync_count__read_is_too_close_behind_write++;

                if (_resyncLoggingEnabled) {
                    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "RESYNC (read fast):  read: %I64d  write: %I64d  min:%I64d  overflow:%I64d",
                        _read_index_distance_traveled,
                        _write_index_distance_traveled,
                        (Sint64)_read_minimum_trailing_distance,
                        (Sint64)_read_minimum_trailing_distance - distance);
                    log_write((char*)_debugBuffer);
                }
            }
        }
        _soundbuffer_rewind_read();
    }
    else if (distance > (Sint64)_read_maximum_trailing_distance) {
        // read is too far behind write (read is too slow)
        if (!timing_is_faster()) {
            // only gather statistics for when CPU is running at normal speed
            if (!_soundbuffer_is_in_initial_sync_period()) {
                // and not in the initial sync period
                _resync_count__read_is_too_far_behind_write++;

                if (_resyncLoggingEnabled) {
                    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "RESYNC (read slow):  read: %I64d  write: %I64d  max:%I64d  underflow:%I64d",
                        _read_index_distance_traveled,
                        _write_index_distance_traveled,
                        (Sint64)_read_maximum_trailing_distance,
                        distance - (Sint64)_read_maximum_trailing_distance);
                    log_write((char*)_debugBuffer);
                }
            }
        }
        _soundbuffer_fastforward_read();
    }

    // perform the actual read
    Uint8 value = _buffer[_read_index++];
    _read_index_distance_traveled++;
    if (_read_index == _bufferSize) {
        _read_index = 0;
    }

    return value;
}

void soundbuffer_write(Uint8 value) {
    _buffer[_write_index++] = value;
    _write_index_distance_traveled++;
    if (_write_index == _bufferSize) {
        _write_index = 0;
    }
}

void soundbuffer_enable_resync_logging() {
    log_write("Logging sound resyncs");
    _resyncLoggingEnabled = 1;
}

void soundbuffer_start(Uint32 sampleRatePerSecond) {
    _compute_sync_divisor();

    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Resync log ignore initial interval: %.2f sec",
        _initialSyncPeriodSeconds);
    log_write((char*)_debugBuffer);

    _sampleRatePerSecond = sampleRatePerSecond;
    _bufferSize = (Uint32) ((double)_sampleRatePerSecond * _bufferSizeSeconds);
    _buffer = (Uint8*)malloc(_bufferSize * sizeof(Uint8*));
    if (_buffer == NULL) {
        log_write("Error: could not allocate memory for the audio circular buffer");
        return;
    }
    memset(_buffer, 0, _bufferSize);
    _read_index = 0;
    _write_index = 0;

    _read_index_distance_traveled = 0;
    _write_index_distance_traveled = 0;

    _resync_count__read_is_too_close_behind_write = 0;
    _resync_count__read_is_too_far_behind_write = 0;

    _compute_sync_values();
}

void soundbuffer_destroy() {
    if (_buffer != NULL) {
        free(_buffer);
        _buffer = NULL;
    }

    // time spent reading
    double seconds = (double)_read_index_distance_traveled / (double)_sampleRatePerSecond;
    
    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Sample rate: %u/sec",
        _sampleRatePerSecond);
    log_write((char*)_debugBuffer);
    
    double bufferSizeMs = (double)_bufferSize / (double)_sampleRatePerSecond * 1000.0f;
    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Buffer size: %.2fms (%u samples)",
        bufferSizeMs,
        _bufferSize);
    log_write((char*)_debugBuffer);

    // resync numbers
    double minTrailingDistanceMs = (double)_read_minimum_trailing_distance / (double)_sampleRatePerSecond * 1000.0f;
    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Resync - minimum trailing distance: %u   (%u samples, %.2fms)",
        _sync__get_min_trailing_distance_factor(),
        _read_minimum_trailing_distance,
        minTrailingDistanceMs);
    log_write((char*)_debugBuffer);

    double maxTrailingDistanceMs = (double)_read_maximum_trailing_distance / (double)_sampleRatePerSecond * 1000.0f;
    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Resync - maximum trailing distance: %u   (%u samples, %.2fms)",
        _sync__get_max_trailing_distance_factor(),
        _read_maximum_trailing_distance,
        maxTrailingDistanceMs);
    log_write((char*)_debugBuffer);

    double rewindMs = (double)_read_sync_rewind_amount / (double)_sampleRatePerSecond * 1000.0f;
    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Resync - rewind amount: %u   (%u samples, %.2fms)",
        _sync__get_rewind_amount_factor(),
        _read_sync_rewind_amount,
        rewindMs);
    log_write((char*)_debugBuffer);

    double fastforwardMs = (double)_read_sync_fastforward_amount / (double)_sampleRatePerSecond * 1000.0f;
    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Resync - fast forward amount: %u   (%u samples, %.2fms)",
        _sync__get_fastforward_amount_factor(),
        _read_sync_fastforward_amount,
        fastforwardMs);
    log_write((char*)_debugBuffer);

    // we don't report resyncs during the sync period
    seconds -= _initialSyncPeriodSeconds;

    double readTooClosePerSecond = 0.0f;
    double readTooFarPerSecond = 0.0f;

    if (seconds > 0.0f) {
        readTooClosePerSecond = (double)_resync_count__read_is_too_close_behind_write / seconds;
        readTooFarPerSecond = (double)_resync_count__read_is_too_far_behind_write / seconds;
    }

    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Sound resync (read too fast): %u (%.1f/sec)",
        _resync_count__read_is_too_close_behind_write,
        readTooClosePerSecond);
    log_write((char*)_debugBuffer);

    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Sound resync (read too slow): %u (%.1f/sec)",
        _resync_count__read_is_too_far_behind_write,
        readTooFarPerSecond);
    log_write((char*)_debugBuffer);
}
